/*---------------------------------------------------------------------------*/
/* g e n e r a t e   --   generate data with which to test an edge detector  */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Standard include files                                                    */
/*---------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pbmplus.h>


/*---------------------------------------------------------------------------*/
/* Symbolic constants                                                        */
/*---------------------------------------------------------------------------*/

#define MAXLINE 200
#define NTABLE 64
#define RANDOM(i) ((unsigned) 69069 * (i) + 1)


/*---------------------------------------------------------------------------*/
/* Local data types                                                          */
/*---------------------------------------------------------------------------*/

typedef enum {FALSE, TRUE} logical;
typedef enum {VERTICAL, HORIZONTAL, PLUS45, MINUS45} direction;
typedef short int pixel_t;


/*---------------------------------------------------------------------------*/
/* Global variables                                                          */
/*---------------------------------------------------------------------------*/

char *progname = "<unknown>";
logical display = FALSE;
char *display_format = "xv %s&";

unsigned shuffle[NTABLE], seed = 632173;
logical unshuffled = TRUE;


/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

void addcon_im (pixel_t **, int, int, int);
void add_noise (pixel_t **, int, int, float);
char * dir_type (direction);
pixel_t **create_im (int, int);
void fill_im (pixel_t **, int, int, direction);
float grng (float, float);
void mult_im (pixel_t **, int, int, int);
void output_im (pixel_t **, int, int, char *);
void set_seed (unsigned);
float urng (float);
void usage (char *);


/*---------------------------------------------------------------------------*/
/* M a i n   p r o g r a m                                                   */
/*---------------------------------------------------------------------------*/

int
main (int argc, char *argv[])
{
  int a;
  logical noisy = FALSE;
  float noise_sd = 0.0f;
  direction edge = VERTICAL;
  logical verbose = FALSE;
  int m = 256, n = 256;
  pixel_t **im;
  pixel_t amp = 1, offset = 0;


  /* Cycle through the command-line arguments, setting variables according */
  /* to the qualifiers supplied. */

  progname = argv[0];
  for (a = 1; a < argc; a++) {
    if (! strcmp(argv[a], "-noisy")) {
      noisy = TRUE;
    } else if (! strcmp(argv[a], "-noise-sd")) {
      a++;
      if (a >= argc)
	usage (progname);
      noise_sd = atof(argv[a]);
    } else if (! strcmp(argv[a], "-size")) {
      a++;
      if (a >= argc)
	usage (progname);
      m = n = atoi(argv[a]);
    } else if (! strcmp(argv[a], "-by")) {
      a++;
      if (a >= argc)
	usage (progname);
      n = atoi(argv[a]);
    } else if (! strcmp(argv[a], "-amplitude")) {
      a++;
      if (a >= argc)
	usage (progname);
      amp = atoi(argv[a]);
    } else if (! strcmp(argv[a], "-offset")) {
      a++;
      if (a >= argc)
	usage (progname);
      offset = atoi(argv[a]);
    } else if (! strcmp(argv[a], "-help")) {
      usage (progname);
    } else if (! strcmp(argv[a], "-vertical")) {
      edge = VERTICAL;
    } else if (! strcmp(argv[a], "-horizontal")) {
      edge = HORIZONTAL;
    } else if (! strcmp(argv[a], "-plus45")) {
      edge = PLUS45;
    } else if (! strcmp(argv[a], "-minus45")) {
      edge = MINUS45;
    } else if (! strcmp(argv[a], "-verbose")) {
      verbose = TRUE;
    } else if (! strcmp(argv[a], "-silent")) {
      verbose = FALSE;
    } else if (argv[a][0] == '-') {
      fprintf (stderr, "%s -- error: unrecognized qualifier \"%s\"\n",
	       progname, argv[a]);
      usage (progname);
    }
  }

  /* Do some miscellaneous checking. */

  if (m < 10 || n < 10) {
    fprintf (stderr, "%s -- error: images must have both dimensions > 10.\n",
	     progname);
  }

  if (noisy && noise_sd <= 0.0) {
    fprintf (stderr, "%s -- error: noise s.d. should be > zero!\n",
	     progname);
    exit (EXIT_FAILURE);
  }

  /* Summarize the options (for debugging). */

  if (verbose) {
    fprintf (stderr, "Image size: %d x %d\n", m, n);
    fprintf (stderr, "Add noise to image: %d\n", noisy);
    fprintf (stderr, "Noise standard deviation: %f\n", noise_sd);
    fprintf (stderr, "Edge direction: %s\n", dir_type(edge));
  }


  /* Generate the image. */

  im = create_im (m, n);
  fill_im (im, m, n, edge);
  mult_im (im, m, n, amp);
  addcon_im (im, m, n, offset);
  if (noisy) {
    add_noise (im, m, n, noise_sd);
  }
  output_im (im, m, n, "-");
  
  return EXIT_SUCCESS;
}

/*---------------------------------------------------------------------------*/
/* u s a g e   --   output a usage message and exit                          */
/*---------------------------------------------------------------------------*/

void
usage (char *p)
{
  fprintf (stderr, "Usage:\n   %s [qualifiers]\nwhere <qualifiers> are:\n",
	   progname);
  fprintf (stderr, "   -help           output this message\n");
  fprintf (stderr, "   -size <n>       number of pixels per image line\n");
  fprintf (stderr, "   -by <n>         number of lines in image\n");
  fprintf (stderr, "   -noisy          add Gaussian noise to image\n");
  fprintf (stderr, "   -noise-sd <n>   s.d. of Gaussian noise to add\n");
  fprintf (stderr, "   -amplitude <n>  amplitude of image\n");
  fprintf (stderr, "   -offset <n>     offset of image\n");
  fprintf (stderr, "   -vertical       insert vertical edge in image\n");
  fprintf (stderr, "   -horizontal     insert horizontal edge in image\n");
  fprintf (stderr, "   -plus45         insert edge in image at +45 degrees\n");
  fprintf (stderr, "   -minus45        insert edge in image at -45 degrees\n");
  fprintf (stderr, "   -verbose        generate messages\n");
  fprintf (stderr, "   -silent         do not generate messages\n");
  exit (EXIT_FAILURE);
}


/*---------------------------------------------------------------------------*/
/* d i r _ t y p e   --   return a string giving the edge direction type     */
/*---------------------------------------------------------------------------*/

char *
dir_type (direction d)
{
  if (d == VERTICAL) {
    return "vertical";
  } else if (d == HORIZONTAL) {
    return "horizontal";
  } else if (d == PLUS45) {
    return "plus 45";
  } else if (d == MINUS45) {
    return "minus 45";
  } else {
    fprintf (stderr, "dir_type -- error: direction type of %d!\n", d);
    exit (EXIT_FAILURE);
  }
}


/*---------------------------------------------------------------------------*/
/* c r e a t e _ i m   --   create an image structure                        */
/*---------------------------------------------------------------------------*/

pixel_t **
create_im (int m, int n)
{
  int i;
  pixel_t **p;

  p = (pixel_t **)malloc(n * sizeof(pixel_t *));
  if (p == (pixel_t **)NULL)
    return p;

  for (i = 0; i < n; i++) {
    p[i] = (pixel_t *)malloc(m * sizeof(pixel_t));
    if( p[i] == (pixel_t *)NULL)
      return (pixel_t **)0;
  }
  return p;
}


/*---------------------------------------------------------------------------*/
/* f i l l _ i m   --   set the image to contain an edge                     */
/*---------------------------------------------------------------------------*/

void
fill_im (pixel_t **im, int m, int n, direction d)
{
  int i, j, t;
  pixel_t v;

  if (d == VERTICAL) {
    t = m / 2 + 1;
    for (j = 0; j < n; j++) {
      for (i = 0; i < m; i++) {
	v = (i < t)? (pixel_t)0 : (pixel_t)1;
	im[j][i] = v;
      }
    }
  } else if (d == HORIZONTAL) {
    t = n / 2 + 1;
    for (j = 0; j < n; j++) {
      v = (j < t)? (pixel_t)0 : (pixel_t)1;
      for (i = 0; i < m; i++) {
	im[j][i] = v;
      }
    }
  } else if (d == PLUS45) {
    for (j = 0; j < n; j++) {
      t = n - j;
      for (i = 0; i < m; i++) {
	v = (i < t)? (pixel_t)0 : (pixel_t)1;
	im[j][i] = v;
      }
    }
  } else if (d == MINUS45) {
    for (j = 0; j < n; j++) {
      for (i = 0; i < m; i++) {
	v = (j < i)? (pixel_t)0 : (pixel_t)1;
	im[j][i] = v;
      }
    }
  } else {
    fprintf (stderr, "fill_im -- error: direction type of %d!\n", d);
    exit (EXIT_FAILURE);
  }
}


/*---------------------------------------------------------------------------*/
/* m u l t _ i m   --   multiply each pixel by a constant                    */
/*---------------------------------------------------------------------------*/

void
mult_im (pixel_t **p, int m, int n, int v)
{
  int i, j;

  for (j = 0; j < n; j++) {
    for (i = 0; i < m; i++) {
      p[j][i] *= v;
    }
  }
}


/*---------------------------------------------------------------------------*/
/* a d d c o n _ i m   --  add a constant to each pixel                      */
/*---------------------------------------------------------------------------*/

void
addcon_im (pixel_t **p, int m, int n, int v)
{
  int i, j;

  for (j = 0; j < n; j++) {
    for (i = 0; i < m; i++) {
      p[j][i] += v;
    }
  }
}


/*---------------------------------------------------------------------------*/
/* a d d _ n o i s e   --   add Gaussian noise to each pixel                 */
/*---------------------------------------------------------------------------*/

void
add_noise (pixel_t **p, int m, int n, float sd)
{
  int i, j;

  for (j = 0; j < n; j++) {
    for (i = 0; i < m; i++) {
      p[j][i] += grng (0.0, sd);
    }
  }
}


/*---------------------------------------------------------------------------*/
/* o u t p u t _ i m   --   output an image                                  */
/*---------------------------------------------------------------------------*/

void
output_im (pixel_t **p, int m, int n, char *f)
{
  char cmdline[MAXLINE];
  FILE *fout;
  int i, j;
  gray *line;
  logical close, dis;

  if (strcmp (f, "-")) {
    dis = FALSE;
    close = TRUE;
    fout = fopen (f, "wb");
    if (fout == (FILE *)NULL) {
      fprintf (stderr, "%s -- fatal: can't open %s\n", progname, f);
      exit (1);
    }
  } else {
    dis = display;
    close = FALSE;
    fout = stdout;
  }

  /* Write the output in PBMPLUS format. */

  pgm_writepgminit (fout, m, n, (gray)255, 0);
  line = pgm_allocrow (m);
  for (j = 0; j < n; j++) {
    for (i = 0; i < m; i++)
      line[i] = (gray)p[j][i];
    pgm_writepgmrow (fout, line, m, (gray)255, 0);
  }
  pgm_freerow (line);
  if (close) {
    fclose (fout);
  }

  /* If the output is to be displayed, invoke the display program. */

  if (dis) {
    fprintf (stderr, "Displaying image `%s'...\n", f);
    fflush (stderr);
    sprintf (cmdline, display_format, f);
    system (cmdline);
  }
}


/*---------------------------------------------------------------------------*/
/* set_seed -- set the seed for the random number generators                 */
/*---------------------------------------------------------------------------*/

void
set_seed (unsigned xseed)
{
  int i;
  unsigned num;

  /* Remember the seed. */

  num = seed = xseed;

  /* Re-fill the shuffle table and mark it as being filled. */

  for( i = 0; i < NTABLE; i++ )
    num = shuffle[i] = RANDOM(num);
  unshuffled = FALSE;
}


/*---------------------------------------------------------------------------*/
/* urng -- generate a uniform random number in the range (0..maxval]         */
/*---------------------------------------------------------------------------*/

float
urng (float maxval)
{
  register int i;
  register unsigned num;

  /* Ensure that the shuffle table has been filled. */

  if( unshuffled ) {
    num = seed;
    for( i = 0; i < NTABLE; i++ )
      num = shuffle[i] = RANDOM(num);
    unshuffled = FALSE;
  }

  /* We wrap the following code in a forever loop so that zero
     values will be discarded. */

  while (TRUE) {

    /* Generate the next pseudo-random number in the sequence. */

    num = seed = RANDOM( seed );

    /* Improve the randomness by shuffling. */

    i = num % NTABLE;
    num = shuffle[i];
    shuffle[i] = seed;

    /* Pull out enough bits to fill the mantissa. */

    num &= (unsigned) 0xffffff;  /* now in the range 0..2^{24}-1 */

    /* Accept the value if non-zero. */

    if( num != 0 ) break;
  }

  /* Convert to floating-point into the right range and return it
     (1,677,215 is 2^{24}-1). */

  return (float) num / 16777215.0 * maxval;
}


/*---------------------------------------------------------------------------*/
/* grng -- generate a Gaussian random number with mean `ave', and sd `sd'    */
/*---------------------------------------------------------------------------*/

float
grng (float ave, float sd)
{
  float u, x, x2;

  while( 1 ) {
    u = urng( 1.0 );
    if( u == 0.0 ) continue;	/* reject zero */
    x = 1.71552777 * (urng(1.0) - 0.5) / u;

    x2 = x * x;
    if( x2 <= (5.0 - 5.136101664*u) ) break;
    if( x2 >= (1.036961043/u + 1.4) ) continue;
    if( x2 > (-4.0 * log(u)) ) continue;
    break;
  }
  return ave + x * sd;    
}
